-- =====================================================================
-- Copyright © 2010-2012 by Cryptographic Engineering Research Group (CERG),
-- ECE Department, George Mason University
-- Fairfax, VA, U.S.A.
-- =====================================================================

library ieee;
use ieee.std_logic_1164.all;   
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all; 
use work.sha3_pkg.all;
use work.groestl_pkg.all;												  

entity groestl_pq_parallel_ppl is
generic (n:integer := GROESTL_DATA_SIZE_SMALL; rom_style : integer := DISTRIBUTED; p_mode : integer := 1; ppl : integer := 2);
port( 						  
	clk				: in std_logic;
	round			: in std_logic_vector(7 downto 0);
	input 			: in std_logic_vector(n-1 downto 0);
   	output 			: out std_logic_vector(n-1 downto 0));
end groestl_pq_parallel_ppl;  		

architecture struct of groestl_pq_parallel_ppl is
	 
	signal round_in 	: std_logic_vector(7 downto 0);
	signal addcons_in 	: std_logic_vector(n-1 downto 0);
	signal subbyte_out 	: std_logic_vector(n-1 downto 0);
	signal shiftout	   	: std_logic_vector(n-1 downto 0);
	signal mixbytes_in 	: std_logic_vector(n-1 downto 0);
	signal mixbytes_out	: std_logic_vector(n-1 downto 0);
	signal addcons		: std_logic_vector(n-1 downto 0);

begin
				
	pmode0Gen: 
		if p_mode = 0 generate
			addconsGen: 
				for i in 0 to n/64-1 generate
					addcons(i*64+63 downto i*64+ 8) <= not addcons_in(i*64+63 downto i*64+ 8);
					addcons(i*64+ 7 downto i*64+ 0) <= (conv_std_logic_vector((16-n/64)+i,4) & x"F" ) xor addcons_in(i*64+ 7 downto i*64+ 0) xor round_in;
				end generate;
			
			srq:
				entity work.groestl_shiftrow(groestl_shiftrowq) 
				generic map (n=>n)
				port map (input=>addcons, output=>shiftout);
		end generate;  
		
	pmode1Gen: 
		if p_mode = 1 generate
			addconsGen: 
				for i in 0 to n/64-1 generate
					addcons(i*64+63 downto i*64+56) <= addcons_in(i*64+63 downto i*64+56) xor round_in xor (conv_std_logic_vector(n/64-1-i,4) & x"0" );
					addcons(i*64+55 downto i*64+ 0) <= addcons_in(i*64+55 downto i*64+ 0);
				end generate;
			
			srp:			
				entity work.groestl_shiftrow(groestl_shiftrowp) 
				generic map (n=>n)
				port map (input=>addcons, output=>shiftout);
		end generate;		
												   						   
	
	sbox_gen: for i in 0 to n/AES_SBOX_SIZE - 1  generate
	sbox	: aes_sbox 	generic map (rom_style=>rom_style)
			port map (	 
				input=>shiftout(AES_SBOX_SIZE*i + 7 downto AES_SBOX_SIZE*i), 
				output=>subbyte_out(AES_SBOX_SIZE*i+7 downto AES_SBOX_SIZE*i));	
	end generate;		  
	
	-- Middle Stage Pipeline Register
	PipelineReg: process( clk )
	begin
		if rising_edge( clk ) then
			mixbytes_in <= subbyte_out;
		end if;
	end process;
			  
	mc:
		entity work.groestl_mixbytes(groestl_mixbytes)	
		generic map (n=>n)
		port map (input=>mixbytes_in,  output=>mixbytes_out);	
	
	PPL2Gen: if PPL = 2 generate
		addcons_in  <= input;
		output 		<= mixbytes_out;
		round_in    <= round;		
	end generate;
	
	PPL4Gen: if PPL = 4 generate
		PipelineReg: process( clk )
		begin
			if rising_edge( clk ) then
				-- First Reg
				addcons_in  <= input;		
				round_in    <= round;		
				
				-- Last Reg
				output 		<= mixbytes_out;
			end if;
		end process;		
	end generate;
	
end struct; 